import asyncio
import requests
import warnings
import webbrowser
from typing import Union

import orjson
import pandas as pd

from prism._utils.exceptions import PrismNotFoundError, PrismResponseError, PrismTypeError, PrismValueError

from ..._common.config import *
from ..._common.const import *
from prism import _core
from ..._utils import (
    _validate_args,
    get,
    post,
    patch,
    delete,
    _authentication,
    _process_fileresponse,
    _process_response,
    _fetch_and_parse,
    _get_web_authentication_token,
    Loader,
    download,
    plot_tree,
    get_sm_attributevalue,
)
from ..._prismcomponent import prismcomponent as pcmpt
from ._universe import parse_universe_to_universeid


__all__ = [
    "list_dataquery",
    "load_dataquery",
    "save_dataquery",
    "rename_dataquery",
    "delete_dataquery",
    "extract_dataquery",
    "get_data",
    "view_data",
]


@_validate_args
def list_dataquery(search: str = None, tree=False):
    dataqueries_df = _fetch_and_parse(URL_DATAQUERIES, "pdq", search)
    if dataqueries_df.empty:
        raise PrismNotFoundError("No Dataqueries found.")
    dataqueries_df = dataqueries_df[dataqueries_df["folderflag"] == 0]
    if not tree:
        return dataqueries_df.drop("folderflag", axis=1)
    plot_tree(dataqueries_df["dataqueryname"].tolist(), "dataquery")


@_validate_args
def load_dataquery(dataquery: Union[str, int]):
    headers = _authentication()
    dataqueryid = parse_dataquery_to_dataqueryid(dataquery)
    res = requests.get(url=URL_DATAQUERIES + f"/{dataqueryid}", headers=headers)
    if not res.ok:
        raise PrismResponseError(res.json()["message"])
    query = res.json()["rescontent"]["data"][0]["dataquerybody"]
    return _dataquery_to_component(query)


def _dataquery_to_component(query: dict):
    component_name = query["component_name"]
    category_component_info = COMPONENT2CATEGORY.get(component_name)
    if category_component_info is None:
        component = pcmpt._PrismComponent(**query)
    else:
        category_name = category_component_info["category"]
        component_fn_name = category_component_info["component"]
        category = getattr(_core, category_name)
        component_fn = getattr(category, component_fn_name)
        component = component_fn(**query["component_args"])
    return component


@_validate_args
def rename_dataquery(old: Union[str, int], new: str):
    should_overwrite, err_msg = should_overwrite_dataquery(new, "renaming")
    if not should_overwrite:
        return err_msg
    dataqueryid = parse_dataquery_to_dataqueryid(old)
    ret = patch(URL_DATAQUERIES + f"/{dataqueryid}", {"newpath": new + ".pdq"}, None)
    print(ret["message"])
    return ret


@_validate_args
def save_dataquery(component, dataqueryname: str):
    should_overwrite, err_msg = should_overwrite_dataquery(dataqueryname, "saving")
    if not should_overwrite:
        return err_msg

    ret = post(URL_DATAQUERIES, {"path": dataqueryname + ".pdq"}, component._query)
    print(f'{ret["message"]}: {ret["result"][0]["resulttype"]} is {ret["result"][0]["resultvalue"]}')
    return ret


@_validate_args
def delete_dataquery(dataquery: Union[str, int]):
    dataqueryid = parse_dataquery_to_dataqueryid(dataquery)
    ret = delete(URL_DATAQUERIES + f"/{dataqueryid}")
    print(f'{ret["message"]}')
    return ret


@_validate_args
def datacomponent_args_info(component_name: str):
    return get(URL_COMPONENT + f"/datacomponent/{component_name}/args")


@_validate_args
def extract_dataquery(component: pcmpt._PrismComponent, return_code=False):
    code = post(URL_DATAQUERIES + '/extract', {'dialect': 'python'}, component._query)
    if return_code:
        return code
    else:
        print(code)


@_validate_args
def get_data(
    component,
    universe: Union[str, int] = None,
    startdate: str = None,
    enddate: str = None,
    shownid: list = None,
    name: Union[str, list] = None,
) -> pd.DataFrame:
    query = []
    cmpts = set()

    def add_cmpts(o):
        cmpts = set()
        if o["component_type"] == "datacomponent":
            cmpts.add(o["component_name"])
        else:
            for c in o["children"]:
                cmpts = cmpts | add_cmpts(c)
        return cmpts

    if not isinstance(component, list):
        component = [component]
    for o in component:
        if isinstance(o, pcmpt._PrismComponent):
            query.append(o._query)
            cmpts = add_cmpts(o._query)
        else:
            raise PrismTypeError(f"Please provide Components into get data.")

    if len(cmpts - set(UniverseFreeDataComponentType)) == 0:
        universeid = None
    else:
        universeid, _ = parse_universe_to_universeid(universe)
    default_shownid = True
    if (shownid is not None) and (len(shownid) == 0):
        shownid = None
        default_shownid = False
    if shownid is not None:
        shownid = [get_sm_attributevalue(a) for a in shownid]
    component_names = set([c._component_name for c in component])
    if (len(component_names - set(ShownidFreeComponents)) == 0) & (shownid is not None):
        print('hello')
        warnings.warn(f'Shownid will be ignored for: {list(component_names & set(ShownidFreeComponents))}')

    params = {
        "universeid": universeid,
        "startdate": startdate,
        "enddate": enddate,
        "shownid": shownid,
        "default_shownid": default_shownid,
        "datanames": name,
        "metadata": False,
        "return_url": True,
    }
    headers = _authentication()
    with Loader("\rFetching Data...") as l:
        res = requests.post(url=URL_DATAQUERIES + "/query", params=params, json=query, headers=headers)
    if res.status_code >= 400:
        return _process_fileresponse(res, "data")
    links = _process_response(res)
    if len(links['urls']) == 0:
        print('The dataframe is empty.')
        return
    else:
        data_links_presigned = [url['file_url'] for url in links['urls']]

    datalinks = requests.post(url=URL_STATIC, json=data_links_presigned, headers=headers)
    datalinks = orjson.loads(datalinks.content)['rescontent']['data']['url']

    ret = []
    for link in datalinks:
        res = requests.get(link, stream=True)
        if res.status_code >= 400:
            return _process_fileresponse(res, "data")

        ret.append(download(res))
    return _process_fileresponse(res, "data", ret)


@_validate_args
def view_data(
    component: Union[pcmpt._PrismComponent, list],
    universe: Union[str, int] = None,
    startdate: str = None,
    enddate: str = None,
    shownid: list = None,
    name: Union[str, list] = None,
):
    query = []
    cmpts = set()

    def add_cmpts(o):
        cmpts = set()
        if o["component_type"] == "datacomponent":
            cmpts.add(o["component_name"])
        else:
            for c in o["children"]:
                cmpts = cmpts | add_cmpts(c)
        return cmpts

    if not isinstance(component, list):
        component = [component]
    for o in component:
        if isinstance(o, pcmpt._PrismComponent):
            query.append(o._query)
            cmpts = add_cmpts(o._query)
        else:
            raise PrismTypeError(f"Please provide Components into get data.")

    if len(cmpts - set(UniverseFreeDataComponentType)) == 0:
        universeid = None
    else:
        universeid, universename = parse_universe_to_universeid(universe)
    if shownid is not None:
        shownid = [get_sm_attributevalue(a) for a in shownid]
    params = {
        "universeid": universeid,
        "startdate": startdate,
        "enddate": enddate,
        "shownid": shownid,
        "view": True,
        "return_url": True,
        "datanames": name,
    }
    headers = _authentication()
    with Loader("\rFetching Data...") as l:
        res = requests.post(url=URL_DATAQUERIES + "/query", params=params, json=query, headers=headers)
        if res.status_code >= 400:
            return _process_response(res)
    links = _process_response(res)['urls']
    if len(links) == 0:
        print('The dataframe is empty.')
        return
    else:
        links = links[0]
    view_data_link = ROOT_EXT_WEB_URL + '/view?'
    for k, v in links.items():
        view_data_link += f'{k}={v}&'
    view_data_link += f'universe={universename.replace("&", "%26")}&startdate={startdate}&enddate={enddate}&'
    web_auth_token = _get_web_authentication_token()
    view_data_link += f'token={web_auth_token}'
    webbrowser.open(view_data_link, new=2)


@_validate_args
def parse_dataquery_to_dataqueryid(dataquery: Union[str, int]):
    dataqueries_df = _fetch_and_parse(URL_DATAQUERIES, "pdq")
    if dataqueries_df.empty:
        raise PrismNotFoundError("No dataqueries Found.")

    if isinstance(dataquery, int):
        dataqueries_df = dataqueries_df[dataqueries_df["dataqueryid"] == dataquery]
    elif isinstance(dataquery, str):
        dataqueries_df = dataqueries_df[dataqueries_df["dataqueryname"] == dataquery]
    else:
        raise PrismValueError("Please provide dataquery name or dataquery id for dataquery.")

    if dataqueries_df.empty:
        raise PrismNotFoundError(f"No dataquery matching: {dataquery}")

    dataqueryid = dataqueries_df["dataqueryid"].values[0]
    return dataqueryid


def should_overwrite_dataquery(dataqueryname, operation):
    dataqueries_df = _fetch_and_parse(URL_DATAQUERIES, "pdq")
    if dataqueries_df.empty:
        return True, None

    if dataqueryname in dataqueries_df["dataqueryname"].to_list():
        overwrite = input(f"{dataqueryname} already exists. Do you want to overwrite? (Y/N) \n")
        if overwrite.lower() == "y":
            return True, None
        elif overwrite.lower() == "n":
            return False, f"Abort {operation} dataquery."
        else:
            return False, "Please provide a valid input."
    else:
        return True, None
