import requests
from typing import Union

from prism._utils.exceptions import PrismNotFoundError, PrismTypeError, PrismValueError

from ..._common.config import *
from ..._common.const import *
from ._list import _list_dataitem_index
from ..._utils import (
    _validate_args,
    post,
    patch,
    delete,
    _process_fileresponse,
    _authentication,
    _fetch_and_parse,
    plot_tree,
)


__all__ = ["list_portfolios", "get_portfolio", "rename_portfolio", "delete_portfolio", "save_index_as_portfolio"]


@_validate_args
def list_portfolio(search: str = None, tree=False):
    portfolio_df = _fetch_and_parse(URL_PORTFOLIOS, "ppt", search)
    if portfolio_df.empty:
        raise PrismNotFoundError("No Portfolios found.")
    if not tree:
        return portfolio_df.drop("folderflag", axis=1)
    plot_tree(portfolio_df["portfolioname"].tolist(), "portfolio")


@_validate_args
def get_portfolio(portfolio, startdate: str = None, enddate: str = None, shownid: list = None):
    portfolioid = parse_portfolios_to_portfolioids([portfolio])[0]
    params = {"startdate": startdate, "enddate": enddate, "shownid": shownid}
    headers = _authentication()
    res = requests.get(URL_PORTFOLIOS + f"/{portfolioid}", params=params, headers=headers)
    return _process_fileresponse(res, "universe", res.content)


@_validate_args
def save_index_as_portfolio(
    dataitemid: int,
    startdate: str = None,
    enddate: str = None,
    portfolioname: str = None,
):
    if portfolioname is None:
        indices = _list_dataitem_index()
        portfolioname = indices[indices["dataitemid"] == dataitemid]["dataitemname"].values[0]

    should_overwrite, err_msg = should_overwrite_portfolio(portfolioname, "saving")
    if not should_overwrite:
        return err_msg
    params = {
        "dataitemid": dataitemid,
        "startdate": startdate,
        "enddate": enddate,
        "path": portfolioname + ".ppt",
    }
    ret = post(URL_PORTFOLIOS + f"/index", params, None)
    print(f'{ret["message"]}: {ret["result"][0]["resulttype"]} is {ret["result"][0]["resultvalue"]}')
    return ret


@_validate_args
def rename_portfolio(old: Union[str, int], new: str):
    should_overwrite, err_msg = should_overwrite_portfolio(new, "renaming")
    if not should_overwrite:
        return err_msg
    portfolioid = parse_portfolios_to_portfolioids([old])[0]
    ret = patch(URL_PORTFOLIOS + f"/{portfolioid}", {"newpath": new + ".ppt"}, None)
    print(ret["message"])
    return ret


@_validate_args
def delete_portfolio(portfolio: Union[str, int]):
    portfolioid = parse_portfolios_to_portfolioids([portfolio])[0]
    ret = delete(URL_PORTFOLIOS + f"/{portfolioid}")
    print(ret["message"])
    return ret


def parse_portfolios_to_portfolioids(portfolios: list):
    portfolios_df = _fetch_and_parse(URL_PORTFOLIOS, "ppt")
    if portfolios_df.empty:
        raise PrismNotFoundError("No portfolio Found.")

    portfolionames = []
    portfolioids = []
    for p in portfolios:
        if isinstance(p, int):
            portfolioids.append(p)
        elif isinstance(p, str):
            portfolionames.append(p)
        else:
            raise PrismValueError("Please provide portfolio name or portfolio id for portfolio.")

    ret_portfolioids = set()
    if len(portfolionames) != 0:
        matching_portfolios_df = portfolios_df[portfolios_df["portfolioname"].isin(portfolionames)]
        if len(matching_portfolios_df) != len(portfolionames):
            not_matched = list(set(portfolionames) - set(matching_portfolios_df["portfolioname"].tolist()))
            raise PrismNotFoundError(f"No portfolio matching: {not_matched}")
        ret_portfolioids.update(matching_portfolios_df["portfolioid"].tolist())

    if len(portfolioids) != 0:
        matching_portfolios_df = portfolios_df[portfolios_df["portfolioid"].isin(portfolioids)]
        if len(matching_portfolios_df) != len(portfolioids):
            not_matched = list(set(portfolioids) - set(matching_portfolios_df["portfolioid"].tolist()))
            raise PrismNotFoundError(f"No portfolio matching: {not_matched}")
        ret_portfolioids.update(matching_portfolios_df["portfolioid"].tolist())
    return list(ret_portfolioids)


def should_overwrite_portfolio(portfolioname, operation):
    portfolios_df = _fetch_and_parse(URL_PORTFOLIOS, "ppt")
    if portfolios_df.empty:
        return True, None

    if portfolioname in portfolios_df["portfolioname"].to_list():
        overwrite = input(f"{portfolioname} already exists. Do you want to overwrite? (Y/N) \n")
        if overwrite.lower() == "y":
            return True, None
        elif overwrite.lower() == "n":
            return False, f"Abort {operation} portfolio."
        else:
            return False, "Please provide a valid input."
    else:
        return True, None
