import requests
import webbrowser
from typing import List, Union

from prism._utils.exceptions import PrismTypeError

from ..._common.config import *
from ..._common.const import *
from ..._core._req_builder._universe import parse_universe_to_universeid
from ..._core._req_builder._portfolio import parse_portfolios_to_portfolioids
from ..._utils import _authentication, get, _process_fileresponse, _validate_args, _get_web_authentication_token
from ..._utils.exceptions import PrismValueError
from ..._prismcomponent import prismcomponent as pcmpt, taskcomponent
from ..._prismcomponent import abstract_prismcomponent


__all__ = [
    "screen",
    "factor_backtest",
    "get_factor_backtest_result",
]


@_validate_args
def screen(
    rule: abstract_prismcomponent._AbstractPrismComponent,
    universe: Union[int, str],
    frequency: str,
    startdate: str = None,
    enddate: str = None,
):
    if isinstance(rule, pcmpt._PrismDataComponent) or FUNCTIONS.get(rule._component_name)["type"] != "logical":
        raise PrismTypeError("screen task only available to boolean operations.")
    universeid, _ = parse_universe_to_universeid(universe)
    return taskcomponent._Screen(
        rule_dataquery=rule._query,
        universeid=int(universeid),
        frequency=frequency,
        startdate=startdate,
        enddate=enddate,
    )


@_validate_args
def factor_backtest(
    factor: abstract_prismcomponent._AbstractPrismComponent,
    universe: Union[int, str],
    frequency: str,
    bins: int,
    startdate: str = None,
    enddate: str = None,
    max_days: int = None,
    rank_method: str = "standard",
):
    if (bins < 2) or (bins > 20):
        PrismValueError('The number of bins should be between 2 and 20')
    if isinstance(factor, list):
        factor = [f.query for f in factor]
    else:
        factor = [factor._query]
    universeid, _ = parse_universe_to_universeid(universe)
    return taskcomponent._FactorBacktest(
        factor_dataquery=factor,
        universeid=int(universeid),
        frequency=frequency,
        bins=bins,
        rank_method=rank_method,
        max_days=max_days,
        startdate=startdate,
        enddate=enddate,
    )


@_validate_args
def strategy_backtest(
    trade: abstract_prismcomponent._AbstractPrismComponent,
    universe: Union[int, str],
    market_impact: str,
    commission_fee: str,
    short_loan_fee: str,
    risk_free_rate: str,
    margin_rate: str,
    cash_interest_rate: str,
    initial_position_type: str,
    initial_position_value: int,
    benchmark: List[Union[int, str]],
    currency: str,
    market_impact_value: float = None,
    commission_fee_value: float = None,
    short_loan_fee_value: float = None,
    risk_free_rate_value: float = None,
    margin_rate_value: float = None,
    cash_interest_rate_value: float = None,
    currency_hedge: bool = False,
    startdate: str = None,
    enddate: str = None,
    unpaid_dividend=None,
):
    assert initial_position_type in ["cash", "portfolio"], "Initial Position Type should be one of: cash, portfolio"
    if not isinstance(benchmark, list):
        benchmark = [benchmark]
    benchmark = parse_portfolios_to_portfolioids(benchmark)
    universeid, _ = parse_universe_to_universeid(universe)
    return taskcomponent._StrategyBacktest(
        trade_dataquery=trade._query,
        universeid=int(universeid),
        market_impact=market_impact,
        market_impact_value=market_impact_value,
        commission_fee=commission_fee,
        commission_fee_value=commission_fee_value,
        short_loan_fee=short_loan_fee,
        short_loan_fee_value=short_loan_fee_value,
        risk_free_rate=risk_free_rate,
        risk_free_rate_value=risk_free_rate_value,
        margin_rate=margin_rate,
        margin_rate_value=margin_rate_value,
        cash_interest_rate=cash_interest_rate,
        cash_interest_rate_value=cash_interest_rate_value,
        initial_position_type=initial_position_type,
        initial_position_value=initial_position_value,
        benchmark=benchmark,
        currency=currency,
        currency_hedge=currency_hedge,
        startdate=startdate,
        enddate=enddate,
        unpaid_dividend=unpaid_dividend,
    )


def get_factor_backtest_result(fbid, data=True, report=False):
    return _get_task_result("factor_backtest", "Factor Backtest", fbid, data, report)


def get_strategy_backtest_result(fbid, data=True, report=False):
    return _get_task_result("strategy_backtest", "Strategy Backtest", fbid, data, report)


@_validate_args
def _get_task_result(tasktype, tasktyperepr, resultid: int, data: bool = True, report: bool = True):
    if (data == False) and (report == False):
        raise PrismValueError("Either data or report should be true.")
    if report:
        # Format Definition:
        # Webclient Address/report/report_type/file_id?token=web_auth_token
        url_get_result = URL_TASK + "/" + tasktype + "/report/" + str(resultid)
        file_id = get(url_get_result)["path"].values[0]

        web_auth_token = _get_web_authentication_token()

        print("Fetching A Link to " + tasktyperepr + " Report...")
        link = ROOT_EXT_WEB_URL + "/report/" + tasktype + "/" + str(file_id)
        print("Link to " + tasktyperepr + " Report:")
        print(link)
        link = link[:-1] + "?token=" + web_auth_token
        webbrowser.open(link, new=2)

    if data:
        url_get_result = URL_TASK + "/" + tasktype + "/result/" + str(resultid)
        print("Fetching " + tasktyperepr + " Result Data...")
        headers = _authentication()
        res = requests.get(url_get_result, headers=headers)
        return _process_fileresponse(res, "task_result", res.content)
