import copy
from xml.etree.ElementTree import TreeBuilder
import requests
from typing import Union

import prism
from prism._utils.exceptions import PrismNotFoundError, PrismResponseError, PrismValueError
from ..._common.config import *
from ..._common.const import *
from ..._core._req_builder import _dataquery
from ..._utils import _validate_args, post, patch, delete, _authentication, _fetch_and_parse, plot_tree


__all__ = [
    "list_taskquery",
    "load_taskquery",
    "save_taskquery",
    "delete_taskquery",
    "rename_taskquery",
    "extract_taskquery",
]


@_validate_args
def list_taskquery(search: str = None, tree=False):
    taskqueries_df = _fetch_and_parse(URL_TASKQUERIES, "ptq", search)
    if taskqueries_df.empty:
        raise PrismNotFoundError("No Taskqueries found.")
    if not tree:
        return taskqueries_df.drop("folderflag", axis=1)
    plot_tree(taskqueries_df["taskqueryname"].tolist(), "taskquery")


@_validate_args
def load_taskquery(taskquery: Union[str, int]):
    headers = _authentication()
    taskqueryid = parse_taskquery_to_taskqueryid(taskquery)
    res = requests.get(url=URL_TASKQUERIES + f"/{taskqueryid}", headers=headers)
    if not res.ok:
        raise PrismResponseError(res.json()["message"])
    query = res.json()["rescontent"]["data"][0]["taskquerybody"]
    component_name = query.pop("component_name")
    component = getattr(prism, component_name)
    component_args = query["component_args"]
    for k in copy.deepcopy(list(component_args.keys())):
        if "_dataquery" in k:
            arg_key = k.split("_dataquery")[0]
            dataquery = _dataquery._dataquery_to_component(component_args.pop(k))
            component_args[arg_key] = dataquery
        if k.endswith("id"):
            component_args[k.split("id")[0]] = component_args.pop(k)
        if k == 'newuniversename':
            component_args.pop('newuniversename')
    result = component(**component_args)
    return result


@_validate_args
def save_taskquery(component, taskqueryname: str):
    should_overwrite, err_msg = should_overwrite_taskquery(taskqueryname, "saving")
    if not should_overwrite:
        return err_msg
    ret = post(
        URL_TASKQUERIES,
        {
            "path": taskqueryname + ".ptq",
        },
        component._query,
    )
    print(f'{ret["message"]}: {ret["result"][0]["resulttype"]} is {ret["result"][0]["resultvalue"]}')
    return ret


@_validate_args
def rename_taskquery(old: Union[str, int], new: str):
    should_overwrite, err_msg = should_overwrite_taskquery(new, "renaming")
    if not should_overwrite:
        return err_msg
    taskqueryid = parse_taskquery_to_taskqueryid(old)
    ret = patch(URL_TASKQUERIES + f"/{taskqueryid}", {"newpath": new + ".ptq"}, None)
    print(ret["message"])
    return ret


@_validate_args
def delete_taskquery(taskquery: Union[str, int]):
    dataqueryid = parse_taskquery_to_taskqueryid(taskquery)
    ret = delete(URL_TASKQUERIES + f"/{dataqueryid}")
    print(ret["message"])
    return ret


@_validate_args
def extract_taskquery(component, return_code=False):
    code = post(URL_TASKQUERIES + '/extract', {'dialect': 'python'}, component._query)
    if return_code:
        return code
    else:
        print(code)


@_validate_args
def parse_taskquery_to_taskqueryid(taskquery: Union[str, int]):
    taskqueries_df = _fetch_and_parse(URL_TASKQUERIES, "ptq")
    if taskqueries_df.empty:
        raise PrismNotFoundError("No Taskquery found.")

    if isinstance(taskquery, int):
        taskqueries_df = taskqueries_df[taskqueries_df["taskqueryid"] == taskquery]
    elif isinstance(taskquery, str):
        taskqueries_df = taskqueries_df[taskqueries_df["taskqueryname"] == taskquery]
    else:
        raise PrismValueError("Please provide taskquery path or taskquery id for taskquery.")

    if taskqueries_df.empty:
        raise PrismNotFoundError(f"No taskquery matching: {taskquery}")

    taskqueryid = taskqueries_df["taskqueryid"].values[0]
    return taskqueryid


def should_overwrite_taskquery(taskqueryname, operation):
    taskqueries_df = _fetch_and_parse(URL_TASKQUERIES, "ptq")
    if taskqueries_df.empty:
        return True, None

    if taskqueryname in taskqueries_df["taskqueryname"].to_list():
        overwrite = input(f"{taskqueryname} already exists. Do you want to overwrite? (Y/N) \n")
        if overwrite.lower() == "y":
            return True, None
        elif overwrite.lower() == "n":
            return False, f"Not {operation} taskquery."
        else:
            return False, "Please provide a valid input."
    else:
        return True, None
