import copy
import requests
from typing import Union

from prism._utils.exceptions import PrismNotFoundError, PrismValueError

from ..._common.config import *
from ..._common.const import *
from ._list import _list_dataitem_index
from ..._utils import (
    _validate_args,
    post,
    patch,
    delete,
    get_sm_attributevalue,
    _authentication,
    _process_response,
    _process_fileresponse,
    _fetch_and_parse,
    plot_tree,
)


__all__ = [
    "list_universe",
    "get_universe",
    "upload_universe",
    "save_index_as_universe",
    "combine_universe",
    "rename_universe",
    "delete_universe",
    "get_universe_template",
    "filter_universe",
]


@_validate_args
def list_universe(search: str = None, tree=False):
    universes_df = _fetch_and_parse(URL_UNIVERSES, "puv", search)
    if universes_df.empty:
        raise PrismNotFoundError("No Universes found.")
    if not tree:
        return universes_df.drop("folderflag", axis=1)
    plot_tree(universes_df["universename"].tolist(), "universe")


@_validate_args
def get_universe(universe, startdate: str = None, enddate: str = None, shownid: list = None, expand: bool = False):
    universeid, _ = parse_universe_to_universeid(universe)
    params = {"startdate": startdate, "enddate": enddate, "shownid": shownid, "expand": expand}
    headers = _authentication()
    res = requests.get(URL_UNIVERSES + f"/{universeid}", params=params, headers=headers)
    return _process_fileresponse(res, "universe", res.content)


@_validate_args
def upload_universe(universefile: str, universename: str):
    universefile_params = {"universefile": open(universefile, "rb")}
    headers = {"Authorization": _authentication()["Authorization"], "Client-Channel": "python-extension"}
    res = requests.post(
        url=URL_UNIVERSES + "/flatfile",
        params={"path": universename + ".puv"},
        files=universefile_params,
        headers=headers,
    )
    ret = _process_response(res)
    print(f'{ret["message"]}: {ret["result"][0]["resulttype"]} is {ret["result"][0]["resultvalue"]}')
    return ret


@_validate_args
def save_index_as_universe(
    dataitemid: int,
    startdate: str = None,
    enddate: str = None,
    universename: str = None,
):
    if universename is None:
        indices = _list_dataitem_index(datacomponent="Index Level")
        universename = indices[indices["dataitemid"] == dataitemid]["dataitemname"].values[0]

    should_overwrite, err_msg = should_overwrite_universe(universename, "combining")
    if not should_overwrite:
        return err_msg
    params = {
        "dataitemid": dataitemid,
        "startdate": startdate,
        "enddate": enddate,
        "path": universename + ".puv",
    }
    ret = post(URL_UNIVERSES + f"/index", params, None)
    print(f'{ret["message"]}: {ret["result"][0]["resulttype"]} is {ret["result"][0]["resultvalue"]}')
    return ret


@_validate_args
def combine_universe(universes: list, newuniversename: str):
    if len(universes) < 2:
        SystemExit("Please provide more than 2 universes to combine.")
    should_overwrite, err_msg = should_overwrite_universe(newuniversename, "saving")
    if not should_overwrite:
        return err_msg
    universeids = parse_universes_to_universeids(universes)
    params = {"universeids": universeids, "path": newuniversename + ".puv"}
    ret = post(URL_UNIVERSES + f"/combine", params, None)
    print(f'{ret["message"]}: {ret["result"][0]["resulttype"]} is {ret["result"][0]["resultvalue"]}')
    return ret


@_validate_args
def rename_universe(old: Union[str, int], new: str):
    should_overwrite, err_msg = should_overwrite_universe(new, "renaming")
    if not should_overwrite:
        return err_msg
    universeid, _ = parse_universe_to_universeid(old)
    ret = patch(URL_UNIVERSES + f"/{universeid}", {"newpath": new + ".puv"}, None)
    print(ret["message"])
    return ret


@_validate_args
def delete_universe(universe: Union[str, int]):
    universeid, _ = parse_universe_to_universeid(universe)
    ret = delete(URL_UNIVERSES + f"/{universeid}")
    print(ret["message"])
    return ret


def get_universe_template():
    res = requests.get(url=URL_UNIVERSES + "/template", headers=_authentication())
    return res.text


@_validate_args
def filter_universe(condition: list, universename: str):
    if len(condition) == 0:
        raise PrismValueError('At least one condition must be provided.')
    condition_copy = copy.deepcopy(condition)
    for i in range(len(condition)):
        assert isinstance(condition[i], dict), 'condition must be list of dicts'
        assert set(condition[i].keys()) - {'operator'} == {'attribute', 'value'}, \
            'Valid arguments are "attribute", "value" and "operator"'
        if not isinstance(condition[i]['value'], list): raise PrismValueError('Value must be a list')
        condition_copy[i]['attribute'] = get_sm_attributevalue(condition_copy[i]['attribute'])
    ret = post(URL_UNIVERSES + '/filter', {'path': universename + '.puv'}, condition_copy)
    print(f'{ret["message"]}: {ret["result"][0]["resulttype"]} is {ret["result"][0]["resultvalue"]}')
    return ret


@_validate_args
def parse_universe_to_universeid(universe: Union[str, int]):
    universes_df = _fetch_and_parse(URL_UNIVERSES, "puv")
    if universes_df.empty:
        raise PrismNotFoundError("No universe found.")

    if isinstance(universe, int):
        universes_df = universes_df[universes_df["universeid"] == universe]
    elif isinstance(universe, str):
        universes_df = universes_df[universes_df["universename"] == universe]
    else:
        raise PrismValueError("Please provide universe name or universe id for universe.")

    if universes_df.empty:
        raise PrismNotFoundError(f"No universe matching: {universe}", True, None)
    universeid = universes_df["universeid"].values[0]
    universename = universes_df["universename"].values[0]

    return universeid, universename


def parse_universes_to_universeids(universes: list):
    universes_df = _fetch_and_parse(URL_UNIVERSES, "puv")
    if universes_df.empty:
        raise PrismNotFoundError("No universe found.")

    universenames = []
    universeids = []
    for u in universes:
        if isinstance(u, int):
            universeids.append(u)
        elif isinstance(u, str):
            universenames.append(u)
        else:
            raise PrismValueError("Please provide universe name or universe id for universe.")

    ret_universeids = set()
    if len(universenames) != 0:
        matching_universes_df = universes_df[universes_df["universename"].isin(universenames)]
        if len(matching_universes_df) != len(universenames):
            not_matched = list(set(universenames) - set(matching_universes_df["universename"].tolist()))
            raise PrismNotFoundError(f"No universe matching: {not_matched}")
        universeids_from_universenames = matching_universes_df["universeid"].tolist()
        ret_universeids.update(universeids_from_universenames)

    if len(universeids) != 0:
        matching_universes_df = universes_df[universes_df["universeid"].isin(universeids)]
        if len(matching_universes_df) != len(universeids):
            not_matched = list(set(universeids) - set(matching_universes_df["universeid"].tolist()))
            raise PrismNotFoundError(f"No universe matching: {not_matched}")
        existing_universeids = matching_universes_df["universeid"].tolist()
        ret_universeids.update(existing_universeids)
    return list(ret_universeids)


def should_overwrite_universe(universename, operation):
    universes_df = _fetch_and_parse(URL_UNIVERSES, "puv")
    if universes_df.empty:
        return True, None

    if universename in universes_df["universename"].to_list():
        overwrite = input(f"{universename} already exists. Do you want to overwrite? (Y/N) \n")
        if overwrite.lower() == "y":
            return True, None
        elif overwrite.lower() == "n":
            return False, f"Not {operation} universe."
        else:
            return False, "Please provide a valid input."
    else:
        return True, None
