import copy

from prism._utils.exceptions import PrismTaskError, PrismValueError

from .._common.config import URL_TASK
from .._common.const import TaskComponentType
from .._core._req_builder._universe import should_overwrite_universe
from .._core._req_builder._portfolio import should_overwrite_portfolio
from .._core._req_builder import _task
from .._prismcomponent.prismcomponent import _PrismTaskComponent
from .._core._req_builder._universe import should_overwrite_universe
from .._utils import _validate_args, Loader, post


class _Screen(_PrismTaskComponent):
    _component_name = TaskComponentType.SCREEN

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @_validate_args
    def run(
        self,
        newuniversename: str,
        jobname: str = None,
        frequency: str = None,
        startdate: str = None,
        enddate: str = None,
    ):
        should_overwrite, err_msg = should_overwrite_universe(newuniversename, "screening")
        if not should_overwrite:
            print(f"{err_msg}")
            return
        component_args = copy.deepcopy(self._query["component_args"])
        universeid = component_args.pop("universeid")
        component_args.update({"universeid": int(universeid)})
        component_args.update({"newuniversepath": newuniversename + ".puv"})
        if frequency is not None:
            component_args["frequency"] = frequency
        if startdate is not None:
            component_args["startdate"] = startdate
        if enddate is not None:
            component_args["enddate"] = enddate
        query = {
            "component_type": self._component_type,
            "component_name": self._component_name,
            "component_args": component_args,
        }

        rescontent = None
        with Loader("Screen Running... ") as l:
            rescontent = post(URL_TASK + "/screen", params={'jobname': jobname}, body=query)

        if rescontent["status"] != "Completed":
            raise PrismTaskError("Screen has failed.")
        print(
            f'{rescontent["message"]}: {rescontent["result"][0]["resulttype"]} is {rescontent["result"][0]["resultvalue"]}'
        )
        return rescontent


class _FactorBacktest(_PrismTaskComponent):
    _component_name = TaskComponentType.FACTOR_BACKTEST

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @_validate_args
    def run(
        self,
        jobname: str = None,
        report: bool = True,
        frequency: str = None,
        bins: int = None,
        rank_method: str = None,
        max_days: int = None,
        startdate: str = None,
        enddate: str = None,
    ):
        component_args = copy.deepcopy(self._query["component_args"])
        universeid = component_args.pop("universeid")
        component_args["universeid"] = int(universeid)
        if frequency is not None:
            component_args["frequency"] = frequency
        if bins is not None:
            component_args["bins"] = bins
        if rank_method is not None:
            component_args["rank_method"] = rank_method
        if max_days is not None:
            component_args["max_days"] = max_days
        if startdate is not None:
            component_args["startdate"] = startdate
        if enddate is not None:
            component_args["enddate"] = enddate
        if (component_args['bins'] < 2) or (component_args['bins'] > 20):
            PrismValueError('The number of bins should be between 2 and 20')
        query = {
            "component_type": self._component_type,
            "component_name": self._component_name,
            "component_args": component_args,
        }

        rescontent = None
        with Loader("Factor Backtest Running... ") as l:
            rescontent = post(URL_TASK + "/factor_backtest", params={'jobname': jobname}, body=query)

        if rescontent["status"] != "Completed":
            raise PrismTaskError("Factor Backtest has failed.")
        print(
            f'{rescontent["message"]}: {rescontent["result"][0]["resulttype"]} is {rescontent["result"][0]["resultvalue"]}'
        )
        return _task._get_task_result(
            "factor_backtest",
            "Factor Backtest",
            int(rescontent["result"][0]["resultvalue"]),
            data=False,
            report=report,
        )


class _StrategyBacktest(_PrismTaskComponent):
    _component_name = TaskComponentType.STRATEGY_BACKTEST

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @_validate_args
    def run(self, portfolioname: str, jobname: str = None, report: bool = True):
        should_overwrite, err_msg = should_overwrite_portfolio(portfolioname, 'constructing')
        if not should_overwrite:
            print(f"{err_msg}")
            return
        component_args = copy.deepcopy(self._query["component_args"])
        universeid = component_args.pop("universeid")
        benchmark = component_args.pop("benchmark")

        market_impact = {
            "model": component_args.pop("market_impact"),
            "model_value": component_args.pop("market_impact_value"),
        }
        commission_fee = {
            "model": component_args.pop("commission_fee"),
            "model_value": component_args.pop("commission_fee_value"),
        }
        short_loan_fee = {
            "model": component_args.pop("short_loan_fee"),
            "model_value": component_args.pop("short_loan_fee_value"),
        }
        risk_free_rate = {
            "model": component_args.pop("risk_free_rate"),
            "model_value": component_args.pop("risk_free_rate_value"),
        }
        margin_rate = {
            "model": component_args.pop("margin_rate"),
            "model_value": component_args.pop("margin_rate_value"),
        }
        cash_interest_rate = {
            "model": component_args.pop("cash_interest_rate"),
            "model_value": component_args.pop("cash_interest_rate_value"),
        }
        initial_position = {
            "position_type": component_args.pop("initial_position_type"),
            "position_value": component_args.pop("initial_position_value"),
        }

        task_params = {
            "portfoliopath": portfolioname + ".ppt",
            "universeid": universeid,
            "trade_value_type": "trade",
            "market_impact": market_impact,
            "commission_fee": commission_fee,
            "short_loan_fee": short_loan_fee,
            "risk_free_rate": risk_free_rate,
            "margin_rate": margin_rate,
            "cash_interest_rate": cash_interest_rate,
            "benchmark": benchmark,
            "initial_position": initial_position,
        }
        task_params.update(component_args)
        query = {
            "component_type": "taskcomponent",
            "component_name": "strategy_backtest",
            "component_args": task_params,
        }

        rescontent = None
        with Loader("Strategy Backtest Running... ") as l:
            rescontent = post(URL_TASK + "/strategy_backtest", params={'jobname': jobname}, body=query)

        if rescontent["status"] != "Completed":
            raise PrismTaskError("Strategy Backtest has failed.")
        for r in rescontent["result"]:
            if r["resulttype"] == "strategybacktestid":
                sb_result = r["resultvalue"]
        print(
            f'{rescontent["message"]}: {rescontent["result"][0]["resulttype"]} is \
                {rescontent["result"][0]["resultvalue"]}, {rescontent["result"][1]["resulttype"]} is \
                    {rescontent["result"][1]["resultvalue"]}'
        )
        return _task._get_task_result("strategy_backtest", "Strategy Backtest", int(sb_result), report)
