import glob
import json
import orjson
import os
import pandas as pd
from prism._utils.exceptions import PrismAuthError, PrismError, PrismResponseError
import pyarrow as pa
import py7zr
import io
import requests
import shutil

import prism
from .._utils import _authentication
from .._common.config import *
from .._common.const import *


def _process_response(res, content=None):
    if res.status_code >= 400:
        if res.status_code == 401:
            raise PrismAuthError(f"Please Login First")

        if res.status_code == 422:
            arg = res.json()["detail"][0]["loc"][-1]
            # enum_list = res.json()["detail"][0]["ctx"]["enum_values"]
            raise PrismResponseError(f'Value at "{arg}" is not a valid enumeration member')

        try:
            err_msg = res.json()["message"]
        except Exception as e:
            if content:
                err_msg = content
            else:
                err_msg = res.content
        raise PrismResponseError(err_msg)
    result = None
    if content:
        resjson = orjson.loads(content)
    else:
        resjson = orjson.loads(res.content)
    restype = resjson["restype"]
    rescontent = resjson["rescontent"]
    if restype == "status":
        result = rescontent
    else:
        data = rescontent.get("data")
        if data is not None:
            if isinstance(data, dict):
                if ("columns" in data.keys()) and (("data" in data.keys())):
                    cols = data.get("columns")
                    data = data.get("data")
                    datecolumns = []
                    cols_to_use = []
                    for c in cols:
                        if c["dtype"] == "datetime64[ns]":
                            datecolumns.append(c["column"])
                        cols_to_use.append(c["column"])
                    result = pd.DataFrame(data, columns=cols_to_use)
                    for c in datecolumns:
                        result[c] = pd.to_datetime(result[c], unit="ns")
                else:
                    result = data
            else:
                result = data
        else:
            result = rescontent
    return result


def _process_fileresponse(res, res_data_type, content=None):
    if res.status_code >= 400:
        try:
            err_msg = res.json()["message"]
        except:
            err_msg = res.content
        raise PrismError(err_msg)
    try:
        order = []
        if (res_data_type == "data"):
            dfs = []
            for c in content:
                obj = pa.ipc.RecordBatchFileReader(c)
                dfs.append(obj.read_pandas())
        else: # from here all zip files are received
            dfs = {}
            with py7zr.SevenZipFile(io.BytesIO(content), 'r') as zip:
                for fname, bio in zip.readall().items():
                    if (".txt" in fname):
                        metadata = bio
                        metadata = json.load(metadata)
                        order = metadata.get("order")
                    else:
                        obj = pa.ipc.RecordBatchFileReader(bio)
                        dfs[fname] = obj.read_pandas()
    except Exception as e:
        raise PrismError("Extension response processing error")

    if res_data_type == "data":
        if order:
            return [dfs[i] for i in order]
    if isinstance(dfs, list):
        return dfs[0] if len(dfs) == 1 else dfs
    elif isinstance(dfs, dict):
        return list(dfs.values())[0] if len(dfs) == 1 else list(dfs.values())


def get(url, params=None, stream=False):
    headers = _authentication()
    res = requests.get(url=url, params=params, headers=headers, stream=stream)
    return _process_response(res)


def post(url, params, body, stream=False):
    headers = _authentication()
    res = requests.post(url=url, params=params, json=body, headers=headers, stream=stream)
    return _process_response(res)


def patch(url, params, body):
    headers = _authentication()
    res = requests.patch(url=url, params=params, json=body, headers=headers)
    return _process_response(res)


def delete(url, params=None):
    headers = _authentication()
    res = requests.delete(url=url, params=params, headers=headers)
    return _process_response(res)


def _fetch_and_parse(url, ext, search=None):
    df = get(url, {"search": search})
    if not df.empty:
        directorytype = FILEEXTENSION[ext]
        df["path"] = df["path"].str.split(prism.username + "/" + directorytype + "/", expand=True).iloc[:, -1]
        df["path"] = df["path"].str.split("." + ext, expand=True).iloc[:, 0]
        df = df.rename(columns={"path": directorytype + "name"})
    return df
