from functools import wraps
import inspect
from multiprocessing.sharedctypes import Value
import re
import sys
import pandas as pd

from prism._utils.exceptions import PrismError, PrismTypeError, PrismValueError
from .._common.const import *


def handle_jupyter_exception(func):
    @wraps(func)
    def showtraceback(*args, **kwargs):
        # extract exception type, value and traceback
        etype, evalue, tb = sys.exc_info()
        customType = "Prism" in etype.__name__
        if customType:
            print(evalue)
            return
        else:
            # otherwise run the original hook
            value = func(*args, **kwargs)
            return value
    return showtraceback


def handle_sys_exception(exc_type, exc_value, exc_traceback):
    customType = "Prism" in exc_type.__name__
    if customType:
        print(exc_value)
        return
    else:
        # otherwise run the original hook
        return sys.__excepthook__(exc_type, exc_value, exc_traceback)


def _validate_args(function):
    @wraps(function)
    def type_checker(*args, **kwargs):
        func_params_dict = inspect.signature(function).parameters

        # construct param-arg pairs: order matters - do not use set operations
        pos_params = tuple(p for p in func_params_dict.keys() if p not in kwargs.keys())
        if "self" in pos_params:
            pos_params = tuple(p for p in pos_params if p != "self")
            pos_args = dict(zip(pos_params, args[1:]))
        else:
            pos_args = dict(zip(pos_params, args))
        all_args = kwargs.copy()
        all_args.update(pos_args)

        args = list(args)
        for param, v in func_params_dict.items():

            # if None, not provided (nor required)
            arg = all_args.get(param, v.default)
            if (arg is not None) and (arg is not inspect._empty):
                # check types
                param_type = v.annotation  # if empty, no hints
                if param_type is not inspect._empty:
                    if hasattr(param_type, "__args__"):
                        unionArgs = param_type.__args__
                        if unionArgs is not None and type(arg) not in unionArgs:
                            raise PrismTypeError(f"Type of {param} is {type(arg)} and not in the Union of {unionArgs}")
                    else:
                        if not issubclass(type(arg), param_type):
                            raise PrismTypeError(f"Type of {param} is {type(arg)} and not {param_type}")

                # validation rules
                arg_enum = None
                if param == "frequency":
                    try:
                        arg_enum = FrequencyType(arg)
                    except Exception as e:
                        raise PrismValueError(f"{e}", valid_list=FrequencyType)
                elif param == "datetype":
                    try:
                        arg_enum = DateType(arg)
                    except Exception as e:
                        raise PrismValueError(f"{e}", valid_list=DateType)
                elif param == "periodtype":
                    try:
                        arg_enum = PeriodType(arg)
                    except Exception as e:
                        raise PrismValueError(f"{e}", valid_list=PeriodType)
                elif param == "adjustment":
                    if arg not in [a.value for a in AdjustmentType]:
                        raise PrismValueError(f"{e}", valid_list=AdjustmentType)
                    # try:
                    #     arg_enum = AdjustmentType(arg)
                    # except Exception as e:
                    #     raise PrismValueError(f"{e}", valid_list=AdjustmentType)
                elif param == "shownid":
                    for idx, id in enumerate(arg):
                        arg[idx] = get_sm_attributevalue(id)
                elif param == "rank_method":
                    try:
                        arg_enum = RankType(arg)
                    except Exception as e:
                        raise PrismValueError(f"{e}", valid_list=RankType)
                elif param in ["startdate", "enddate"]:
                    try:
                        date = pd.to_datetime(arg)
                    except Exception as e:
                        raise PrismValueError(f'Unknown string format. Cannot parse "{arg}" to a date')

                    assert (date >= BEGINNING_DATE) & (date <= ACTIVE_DATE), "Not a valid date."

                elif "universename" in param:
                    regex = re.compile("[@_!#$%^&*()<>?/\|}{~:]`\"'")
                    if isinstance(arg, list):
                        for a in arg:
                            if regex.search(a) is not None:
                                raise PrismValueError("universename not to include special characters")
                    else:
                        if regex.search(arg) is not None:
                            raise PrismValueError("universename not to include special characters")
                elif (function.__name__ == "log") & (param == "base"):
                    if arg <= 0:
                        raise Exception("base condition error")

                if arg_enum is not None:
                    if param in pos_args.keys():
                        for i, k in enumerate(func_params_dict.keys()):
                            if param == k:
                                break
                        args[i] = arg_enum
                    else:
                        kwargs[param] = arg_enum

        return function(*args, **kwargs)

    return type_checker


def get_sm_attributevalue(attribute: str):
    from .._common.const import SMValues
    smattributes_lower = {
        a.lower().replace(' ', '').replace('_', ''): SMValues[a]
        for a in SMValues.keys()
    }
    smattributes_lower.update({
        a.lower().replace(' ', '').replace('_', ''): a
        for a in SMValues.values()
    })
    attribute_lower = attribute.lower().replace(' ', '').replace('_', '')
    smattributevalue = smattributes_lower.get(attribute_lower)
    if smattributevalue is None:
        raise PrismValueError(
            f"{attribute} is not a valid Security Master attribute",
            valid_list=list(SMValues.keys())
        )
    return smattributevalue